#include "xml-element.hpp"
#include "xml-parser.hpp"
#include "xml-handler.hpp"
#include "types.hpp"

#include <vector>
#include <map>
#include <algorithm>
#include <iostream>
#include <fstream>
#include <string>
#include <getopt.h>

using simplexml::XMLElement;
using simplexml::XMLHandler;
using simplexml::XMLParser;
using simplexml::XMLException;

using std::cout;
using std::endl;
using std::cerr;

using std::find_if;
using std::find;

using std::string;
using std::vector;
using std::map;
using std::ofstream;


static struct option LongOptions[] = {
	{"help",    0, 0, 'h'},
	{"version", 0, 0, 'V'},
	{"input",  1, 0, 'i'},
	{"output",  1, 0, 'o'},
	{0,  0, 0, 0}
};

CCHAR_P PATH_SEP = "/";

namespace
{

void PrintHelp()
{
	cout << "Usage: simplexml_xsg -i <XMLFILE> -o <HPPFILE>" << endl
	     << "Options:" << endl
	     << "    -i <XMLFILE>, --input=<XMLFILE>  get specification from XMLFILE"
	     << endl
	     << "    -o <HPPFILE>, --output=<HPPFILE> save result code as HPPFILE"
	     << endl
	     << "    -h, --help                       print this message and exit"
         << endl
	     << "    -V, --version                    print version and exit"
         << endl;
}


void PrintVersion()
{
	cout << "SIMPLEXML XML config scheme generator "
	     << VERSION << endl
	     << "Copyright (C) 2009 Rambler Internet Holding" << endl;

}


class IsNotWord : public std::unary_function<CHAR_8, bool>
{
public:
	bool operator()(CHAR_8 sym)
	{
		if (!isdigit(static_cast<INT_32>(sym)) &&
		    !isalpha(static_cast<INT_32>(sym)) &&
		    sym != '_')
		{
			return true;
		}
		return false;
	}
};


bool CheckWord(const string &word)
{
	string::const_iterator it =
		find_if(word.begin(), word.end(), IsNotWord());

	if (it != word.end()) { return false; }

	return true;
}


void GenHeaderCode(const string &name, const vector<string> &namespaces,
		   ofstream &out)
{
	out << "#ifndef SIMPLEXML_GENERATED_" << name;
	out << "\n";
	out << "#define SIMPLEXML_GENERATED_" << name;
	out << "\n";
	out << "\n";
	out << "/******************************************************************************\n";
	out << " *WARNING: DON'T EDIT THIS! THE FILE GENERATED FROM XML!\n";
	out << " ******************************************************************************/\n";
	out << "\n";
	out << "#include <simplexml/xml-element.hpp>\n";
	out << "#include <simplexml/xml-validator.hpp>\n";
	out << "\n";

	vector<string>::const_iterator it = namespaces.begin();
	for (; it != namespaces.end(); ++it)
	{
		out << "namespace " << *it;
		out << "\n{\n\n";
	}

	out << "class " << name << " : public simplexml::XMLValidator";
	out << "\n{\n";
	out << "public:\n";
	out << "\n";
	out << "\tvirtual const simplexml::XMLElement *SpecRoot() const throw() { return specRoot; }\n";
	out << "\n\t";

	out << "virtual ~" << name << "() throw()";
	out << "\n\t{\n";
	out << "\t\tif (specRoot != NULL) { delete specRoot; }\n";
	out << "\t}\n";
	out << "\n\t";

	out << name << "():";
	out << "\n\t\tspecRoot(NULL)\n";
	out << "\t{\n";
	out << "\t\tsimplexml::XMLElement *elem = 0;\n";
	out << "\t\tsimplexml::XMLElement *curr = new simplexml::XMLElement(elem, \"specification\", 0, 0);\n";
	out << "\t\tcurr->SetAttr(\"name\", \"specification\");\n";
	out << "\t\tstd::vector<simplexml::XMLElement *> levels;\n";
	out << "\t\tspecRoot = curr;\n\n";

}


void GenFooterCode(const string &name, const vector<string> &namespaces,
		   ofstream &out)
{
	out << "\t}\n";
	out << "\n";
	out << "private:\n";
	out << "\n";
	out << "\tsimplexml::XMLElement *specRoot;\n";
	out << "\n";
	out << "}; // class " << name;
	out << "\n\n";

	vector<string>::const_reverse_iterator it = namespaces.rbegin();
	for (; it != namespaces.rend(); ++it)
	{
		out << "} // namespace " << *it;
		out << "\n\n";
	}
	out << "#endif // _SIMPLEXML_GENERATED_" << name << "__";
	out << "\n\n";
}


bool ProcessAttribute(const XMLElement *node, ofstream &out)
{
	string attrType("string");
	if (node->HasAttr("type"))
	{
		attrType = node->GetAttr("type").AsString();
	}

	if (attrType != "string" &&
	    attrType != "integer" &&
	    attrType != "float" &&
	    attrType != "option")
	{
		cerr << "No such type: " << attrType << endl;
		return false;
	}

	out << "\t\tcurr->SetAttr(\"type\", \"" << attrType << "\");\n";

	string defaultValue("_empty_");
	if (node->HasAttr("default"))
	{
		defaultValue = node->GetAttr("default").AsString();
		if (node->Name() == "attribute" && node->HasChildren("option"))
		{
			vector<string> options;
			UINT_32 optionsMax = node->HasChildren("option");
			for (UINT_32 i = 0; i < optionsMax; ++i)
			{
				const XMLElement *xOption =
					node->Child("option", i);

				if (!xOption->HasAttr("name"))
				{
					cerr << "No `name' attribute for `option'" << endl;
					return false;
				}

				options.push_back(xOption->GetAttr("name").AsString());
			}

			vector<string>::const_iterator it =
				find(options.begin(), options.end(),
				     defaultValue);
			if (it == options.end())
			{
				cerr << "Invalid default value" << endl;
				return false;
			}
		}
	}

	out << "\t\tcurr->SetAttr(\"default\", \"" << defaultValue << "\");\n";

	if (node->HasAttr("min"))
	{
		string minAttr = node->GetAttr("min").AsString();
		out << "\t\tcurr->SetAttr(\"min\", \"" << minAttr << "\");\n";
	}

	if (node->HasAttr("max"))
	{
		string maxAttr = node->GetAttr("max").AsString();
		out << "\t\tcurr->SetAttr(\"max\", \"";
		out << maxAttr << "\");\n";
	}

	return true;

}

bool ProcessElement(const XMLElement *node, ofstream &out)
{
	map<string, string> elements;
	elements["repeated"] = "0";
	elements["required"] = "1";
	elements["checked"] = "1";

	map<string, string>::iterator it = elements.begin();

	for (; it != elements.end(); ++it)
	{
		if (node->HasAttr(it->first))
		{
			it->second = node->GetAttr(it->first).AsString();
		}

		if (it->second != "0" &&it->second != "1")
		{
			cerr << "attribute `" << it->first;
			cerr << "' must be `0' or `1'" << endl;
			return false;
		}

		out << "\t\tcurr->SetAttr(\"" << it->first << "\", \"";
		out << it->second << "\");\n";
	}

	return true;
}


bool ProcessOption(const XMLElement */*node*/, ofstream &/*out*/,
                   const XMLElement *parent = NULL)
{
	if (parent->Name() != "attribute")
	{
		cerr << "`option' permitted only in attribute, but founded in ";
		cerr << parent->Name();
		return false;
	}


	return true;
}

bool GenElements(const XMLElement *node, ofstream &out,
                 const XMLElement *parent = NULL)
{
	string tagname = node->Name();

	if (tagname != "element" &&
	    tagname != "attribute" &&
	    tagname != "option")
	{
		cerr << "Unknown tag `" << tagname << "'" << endl;
		return false;
	}

//      TODO: FIX: tagname = node.tag.lower()

	if (!node->HasAttr("name"))
	{
		cerr << "No required attribute `name'" << endl;
		return false;
	}

	string name = node->GetAttr("name").AsString();


	out << "\t\telem = new simplexml::XMLElement(curr, \"" << tagname << "\", 0, 0);\n";
	out << "\t\tlevels.push_back(curr);\n";
	out << "\t\tcurr->AddChild(elem);\n";
	out << "\t\tcurr = elem;\n";
	out << "\t\tcurr->SetAttr(\"name\", \"" << name << "\");\n";

	if (tagname == "attribute")
	{
		if (!ProcessAttribute(node, out)) { return false; }
	}
	else if (tagname == "element")
	{
		if (!ProcessElement(node, out)) { return false; }
	}
	else if (tagname == "option")
	{
		if (!ProcessOption(node, out, parent)) { return false; }
	}

	const XMLElement::ChildList childs = node->Children();
	XMLElement::ChildList::const_iterator it = childs.begin();
	for (; it != childs.end(); ++it)
	{
		if (!GenElements(*it, out, node)) { return false; }
	}

	out << "\t\tcurr = levels.back();\n";
	out << "\t\tlevels.pop_back();\n";

	return true;
}

bool GenerateHPP(const string &inputFile, const string &outputFile)
{
	try
	{
		XMLHandler handler;
		XMLParser parser(&handler);
		parser.ParseFile(inputFile);

		const XMLElement *root = handler.Root();

		if (root->HasChildren("settings") > 1)
		{
			cerr << "Found duplicate `settings'" << endl;
			return false;
		}

		const XMLElement *xSettings = root->Child("settings");
		if (!xSettings->HasAttr("class"))
		{
			cerr << "No `class' attribute for `settings'" << endl;
			return false;
		}
		string clsName = xSettings->GetAttr("class").AsString();

		if (clsName.empty())
		{
			cerr << "No `class' attribute for `settings'" << endl;
			return false;
		}

		if (!CheckWord(clsName))
		{
			cerr << "`class' attribute must be word" << endl;
			return false;
		}


		vector<string> namespaces;
		UINT_32 namespacesMax = xSettings->HasChildren("namespace");
		for (UINT_32 i = 0; i < namespacesMax; ++i)
		{
			const XMLElement *xNamespace =
				xSettings->Child("namespace", i);

			if (!xNamespace->HasAttr("name"))
			{
				cerr << "No `name' attribute for `namespace'" << endl;
				return false;
			}

		    namespaces.push_back(xNamespace->GetAttr("name").AsString());
		}

		ofstream out(outputFile.c_str());
		GenHeaderCode(clsName, namespaces, out);

		UINT_32 elementsMax = root->HasChildren("elements");
		if (elementsMax > 1)
		{
			cerr << "Found duplicate `elements'" << endl;
			return false;
		}

		const XMLElement *xElements = root->Child("elements");
		const XMLElement::ChildList childs = xElements->Children();
		XMLElement::ChildList::const_iterator it = childs.begin();
		for (; it != childs.end(); ++it)
		{
			if (!GenElements(*it, out)) { return false; }
		}

		GenFooterCode(clsName, namespaces, out);

		out.close();

	}
	catch (XMLException &e)
	{
		cerr << e.what() << endl;
		return false;
	}


	return true;
}

} // namespace

INT_32 main(INT_32 argc, CHAR_P argv[])
{
	CCHAR_P helpWarning = "Use `--help' option";

	if (argc == 1)
	{
		cerr << helpWarning << endl;
		exit(EXIT_FAILURE);
	}

	string inputFile;
	string outputFile;
	INT_32 ch = 0;
	while ((ch = getopt_long(argc, argv, "hVi:o:", LongOptions, NULL))
	       != -1)
	{
		switch (ch)
		{
		case 'h':
			PrintHelp();
			exit(EXIT_SUCCESS);
		case 'V':
			PrintVersion();
			exit(EXIT_SUCCESS);
		case 'i':
			inputFile = optarg;
			break;
		case 'o':
			outputFile = optarg;
			break;
		default:
			cerr << helpWarning << endl;
			exit(1);
		}
	}

	if (optind < argc)
	{
		cerr << helpWarning << endl;
		exit(1);
	}

	if (inputFile.empty() || outputFile.empty())
	{
		cerr << helpWarning << endl;
		exit(1);
	}

	if (!GenerateHPP(inputFile, outputFile)) { exit(EXIT_FAILURE); }

	return 0;
}

