// Copyright (c) 2010 Volvox Development Team
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//
// Author: Konstantin Lepa <konstantin.lepa@gmail.com>

#include "xml-element.hpp"
#include "xml-utils.hpp"

#include <locale>
#include <functional>
#include <string>
#include <vector>
#include <map>
#include <algorithm>

#ifndef __APPLE__
static std::locale loc("C");
#else
#warning "ATTENTION: std::locale loc("") can't work under MacOS X (segfault)"
#endif

using std::find;

namespace simplexml
{

XMLElement::XMLElement(XMLElement *parent, const std::string &name,
                       UINT_32 line, UINT_32 column) :
	parent_(parent),
	name_(name),
	line_(line),
	column_(column),
	attrs_(),
	children_()
{
#ifndef __APPLE__
	std::transform(name_.begin(), name_.end(), name_.begin(),
	               std::bind2nd(ToLower(), loc));
#endif
}


XMLElement * XMLElement::Parent() const { return parent_; }


const std::string &XMLElement::Name() const { return name_; }


bool XMLElement::HasAttr(const std::string &attName) const
{
	std::string lowerName;
#ifndef __APPLE__
	std::transform(attName.begin(), attName.end(),
	               std::back_inserter(lowerName),
	               std::bind2nd(ToLower(), loc));
#else
	lowerName = attName;
#endif

	return (attrs_.find(lowerName) != attrs_.end());
}

bool XMLElement::HasDefaultAttr(const std::string &attName) const
{
	std::string lowerName;
#ifndef __APPLE__
	std::transform(attName.begin(), attName.end(),
	               std::back_inserter(lowerName),
	               std::bind2nd(ToLower(), loc));
#else
	lowerName = attName;
#endif

	return (find(defaultAttrs_.begin(), defaultAttrs_.end(),
	             lowerName) != defaultAttrs_.end());
}

const XMLValue &XMLElement::GetAttr(const std::string &attName) const
{
	std::string lowerName;
#ifndef __APPLE__
	std::transform(attName.begin(), attName.end(),
	               std::back_inserter(lowerName),
	               std::bind2nd(ToLower(), loc));
#else
	lowerName = attName;
#endif

	const AttrMap::const_iterator attr(attrs_.find(lowerName));
	if (attr == attrs_.end())
	{
		throw simplexml::XMLException(std::string("no such attribute ") + attName +
		                          std::string(" for ") + Name(),
		                          line_, column_);
	}
	return attr->second;
}

void XMLElement::SetAttr(const std::string &attName, const std::string &value,
                         bool hasDefault)
{
	std::string lowerName;
#ifndef __APPLE__
	std::transform(attName.begin(), attName.end(),
	               std::back_inserter(lowerName),
	               std::bind2nd(ToLower(), loc));
#else
	lowerName = attName;
#endif
	attrs_[lowerName] = value;
	if (hasDefault) { defaultAttrs_.push_back(lowerName); }
}


const simplexml::XMLElement::AttrMap &XMLElement::Attrs() const { return attrs_; }


UINT_32 XMLElement::HasChildren(const std::string &name) const
{
	if (!name.empty())
	{
		return std::count_if(children_.begin(), children_.end(),
		                     std::bind2nd(XMLElementFinder(), name));
	}

	return children_.size();
}


const simplexml::XMLElement::ChildList &XMLElement::Children() const
{
	return children_;
}


XMLElement *XMLElement::Child(const std::string &name, UINT_32 num) const
{
	ChildList::const_iterator it = children_.begin();
	for (UINT_32 i = 0; i < num; ++i, ++it)
	{
		it = std::find_if(it, children_.end(),
		                  std::bind2nd(XMLElementFinder(), name));

		if (it == children_.end())
		{
			throw simplexml::XMLException(std::string("no such number "
			                                      "of children ") +
			                          name);
		}
	}

	if (it != children_.end())
	{
		return *it;
	}

	return NULL;
}


void XMLElement::AddChild(XMLElement *newChild)
{
	children_.push_back(newChild);
}


XMLElement * XMLElement::Child(const std::string &name) const
{
	ChildList::const_iterator it =
		std::find_if(children_.begin(), children_.end(),
		             std::bind2nd(XMLElementFinder(), name));

	if (it != children_.end())
	{
		return *it;
	}
	else
	{
		throw simplexml::XMLException(std::string("no such tag ") + name,
		                          line_, column_);
	}

	return NULL;
}


UINT_32 XMLElement::LineInFile() const { return line_; }


UINT_32 XMLElement::ColumnInFile() const { return column_; }


XMLElement::~XMLElement() throw()
{
	const ChildListSize childCount = children_.size();
	for (ChildListSize i = 0; i < childCount; ++i)
	{
		delete children_[i];
	}
}


} // namespace simplexml

