// Copyright (c) 2010 Volvox Development Team
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//
// Author: Konstantin Lepa <konstantin.lepa@gmail.com>

#include "xml-validator.hpp"
#include "xml-element.hpp"
#include "xml-utils.hpp"

#include <cstdio>
#include <algorithm>
#include <functional>

using simplexml::XMLValidatorException;
using simplexml::XMLElement;

using std::string;

namespace
{

class XMLFindUnknownTag :
	public std::binary_function<const XMLElement *,
	                            const simplexml::XMLValidator::ChildVec,
	                            bool>
{
public:
	bool operator()(const XMLElement *elem,
	                const simplexml::XMLValidator::ChildVec &specChilds) const
	{
		simplexml::XMLValidator::ChildCIter it =
			std::find_if(specChilds.begin(), specChilds.end(),
				     std::bind2nd(simplexml::XMLSpecElementFinder(),
						  elem->Name()));

		if (it == specChilds.end()) { return true; }

		return false;
	}
}; // class XMLFindUnknownTag


class XMLFindUnknownAttr :
	public std::binary_function<const std::pair<string, simplexml::XMLValue>,
	                            const simplexml::XMLValidator::ChildVec,
	                            bool>
{
public:
	bool operator()(const std::pair<string, simplexml::XMLValue> &attr,
	                const simplexml::XMLValidator::ChildVec &specChilds) const
	{
		simplexml::XMLValidator::ChildCIter it =
			std::find_if(specChilds.begin(), specChilds.end(),
				     std::bind2nd(simplexml::XMLSpecElementFinder(),
			                          attr.first));

		if (it == specChilds.end()) { return true; }

		return false;
	}
}; // class XMLFindUnknownAttr

} // anonymous namespace


namespace simplexml
{

XMLValidator::~XMLValidator() throw() { ;; }


void XMLValidator::Validate(XMLElement *root) const
{
	const XMLElement *spec = SpecRoot();
	if (spec == NULL) { return; }

	DiffTag(spec, root);
}


//////////////////////////////////////////////////////////////////////////////
// Private methods
//////////////////////////////////////////////////////////////////////////////

void XMLValidator::AssertNoUnknownTags(const XMLElement *spec,
                                       XMLElement *root) const
{
	const ChildVec &spChilds = spec->Children();
	const ChildVec &childs = root->Children();

	ChildCIter it =
		std::find_if(childs.begin(), childs.end(),
		             std::bind2nd(XMLFindUnknownTag(), spChilds));

	if (it != childs.end())
	{
		throw XMLValidatorException("found unknown element",
		                                 (*it)->Name(),
		                                 (*it)->LineInFile(),
		                                 (*it)->ColumnInFile());
	}
}


void XMLValidator::AssertNoUnknownAttrs(const XMLElement *spec,
                                        XMLElement *root) const
{
	const XMLElement::AttrMap &attrs = root->Attrs();
	const ChildVec &spChilds = spec->Children();


	XMLElement::AttrMap::const_iterator it = attrs.begin();

	for (; it != attrs.end(); ++it)
	{
		simplexml::XMLValidator::ChildCIter spIt =
			std::find_if(spChilds.begin(), spChilds.end(),
				     std::bind2nd(XMLSpecElementFinder(),
						  it->first));

		if (spIt == spChilds.end())
		{
			throw XMLValidatorException("found unknown attribute",
			                                 (*it).first,
							 root->LineInFile(),
							 root->ColumnInFile());
		}

	}

}


bool XMLValidator::FindUnknownOption(const XMLElement *spec,
                                     const string &option) const
{
	typedef std::vector<XMLElement *> ChildVec;
	typedef ChildVec::const_iterator ChildCIter;

	const ChildVec &childs = spec->Children();

	ChildCIter it = childs.begin();
	for (; it != childs.end(); ++it)
	{
		string name((*it)->GetAttr("name").AsString());

		if (name == option) { return false; }
	}

	return true;
}



void XMLValidator::DiffAttr(const XMLElement *spec,
                            XMLElement *root,
                            const string &missingPath) const
{
	typedef std::vector<XMLElement *> ChildVec;
	typedef ChildVec::const_iterator ChildCIter;

	const ChildVec &spChilds = spec->Children();

	AssertNoUnknownAttrs(spec, root);

	ChildVec spAttrVec;
	std::remove_copy_if(spChilds.begin(), spChilds.end(),
	                    std::back_inserter(spAttrVec),
	                    std::not1(std::bind2nd(XMLElementFinder(),
	                              string("attribute"))));

	ChildCIter spIt = spAttrVec.begin();
	for (; spIt != spAttrVec.end(); ++spIt)
	{
		string name((*spIt)->GetAttr("name").AsString());
		string type((*spIt)->GetAttr("type").AsString());
		simplexml::XMLValue dvalue((*spIt)->GetAttr("default").AsString());
		bool isDefault = true;

		const simplexml::XMLValue *val = &dvalue;
		if (root->HasAttr(name))
		{
			val = &root->GetAttr(name);
			isDefault = false;
		}
		else if (dvalue.AsString() == "_empty_")
		{
			std::string msg;
			if (!missingPath.empty())
			{
				msg.append("tag " + missingPath +
				           " doesn't exists, ");
			}

			throw XMLValidatorException(msg + "no value for "
			                            "required attribute",
			                            name,
			                            root->LineInFile(),
			                            root->ColumnInFile());
		}

		try
		{
			simplexml::XMLValue max;
			if ((*spIt)->HasAttr("max"))
			{
				max = (*spIt)->GetAttr("max");
			}

			simplexml::XMLValue min;
			if ((*spIt)->HasAttr("min"))
			{
				min = (*spIt)->GetAttr("min");
			}

			if (type == "float")
			{
				W_FLOAT v = val->AsFloat();
				if (!max.Empty() && v > max.AsFloat())
				{
					throw XMLException("value bigger then max value");
				}

				if (!min.Empty() && v < min.AsFloat())
				{
					throw XMLException("value lesser then max value");
				}

			}
			else if (type == "integer")
			{
				INT_64 v = val->AsInteger();
				if (!max.Empty() && v > max.AsInteger())
				{
					throw XMLException("value bigger then max value");
				}

				if (!min.Empty() && v < min.AsInteger())
				{
					throw XMLException("value lesser then max value");
				}
			}


		}
		catch (simplexml::XMLException &e)
		{
			string entity("");
			if (isDefault) { entity.append("default "); }
			entity.append("attribute ");
			entity.append(name);
			entity.append(" with incorrect type");

			throw XMLValidatorException(e.what(),
			                                 entity,
			                                 root->LineInFile(),
			                                 root->ColumnInFile());
		}

		if (type == "option" && FindUnknownOption(*spIt, val->AsString()))
		{
			throw XMLValidatorException("no such option",
			                                 val->AsString(),
			                                 root->LineInFile(),
			                                 root->ColumnInFile());
		}

		if (isDefault)
		{
			root->SetAttr(name, dvalue.AsString(), true);
		}
	}

}


void XMLValidator::DiffTag(const XMLElement *spec,
                           XMLElement *root,
                           const string &missingPath) const
{
	typedef std::vector<XMLElement *> ChildVec;
	typedef ChildVec::const_iterator ChildCIter;

	const ChildVec &spChilds = spec->Children();
	const ChildVec &childs = root->Children();

	ChildVec spTagVec;
	std::remove_copy_if(spChilds.begin(), spChilds.end(),
	                    std::back_inserter(spTagVec),
	                    std::not1(std::bind2nd(XMLElementFinder(),
	                              string("element"))));


	AssertNoUnknownTags(spec, root);

	ChildCIter spIt = spTagVec.begin();
	for (; spIt != spTagVec.end(); ++spIt)
	{
		string tagName((*spIt)->GetAttr("name").AsString());
		bool repeated = (*spIt)->GetAttr("repeated").AsInteger() == 1;
		bool required = (*spIt)->GetAttr("required").AsInteger() == 1;
		bool checked = (*spIt)->GetAttr("checked").AsInteger() == 1;

		ChildVec tagVec;
		std::remove_copy_if(childs.begin(), childs.end(),
				    std::back_inserter(tagVec),
	                            std::not1(std::bind2nd(XMLElementFinder(),
	                                      tagName)));

		if (!repeated && tagVec.size() > 1)
		{
			throw XMLValidatorException("found duplicated element",
			                            tagName,
			                            tagVec[1]->LineInFile(),
			                            tagVec[1]->ColumnInFile());
		}

		const string elemName = (*spIt)->GetAttr("name").AsString();
		if (required && tagVec.size() == 0)
		{
			XMLElement *elem =
				new XMLElement(root, elemName, 0, 0);
			root->AddChild(elem);
			tagVec.push_back(elem);
		}

		ChildCIter it = tagVec.begin();
		for (; it != tagVec.end(); ++it)
		{
			string newPath = missingPath + "/" + elemName;
			DiffAttr(*spIt, *it, newPath);

			if (checked) { DiffTag(*spIt, *it, newPath); }
		}
	}
}

} // namespace simplexml

