// Copyright (c) 2010 Volvox Development Team
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//
// Author: Konstantin Lepa <konstantin.lepa@gmail.com>

#include "xml-value.hpp"

#include <string>
#include <ctype.h>
#include <errno.h>
#include <string.h>
#include <algorithm>
#include <functional>

namespace
{
class IsNotDigitOrPoint : public std::unary_function<CHAR_8, bool>
{
public:
	bool operator()(CCHAR_8 sym)
	{
		INT_32 c = static_cast<INT_32>(sym);
		if (!isdigit(c) &&c != '.')
		{
			return true;
		}
		return false;
	}
};

class IsNotDigit : public std::unary_function<CHAR_8, bool>
{
public:
	bool operator()(CHAR_8 sym)
	{
		if (!isdigit(static_cast<INT_32>(sym)))
		{
			return true;
		}
		return false;
	}
};

} // namespace

namespace simplexml
{
const std::string & XMLValue::AsString() const throw() { return value_; }


INT_64 XMLValue::AsInteger() const
{
	std::string::const_iterator it = value_.begin();

	if (*it == '-') { ++it; }

	it = std::find_if(it, value_.end(), IsNotDigit());

	if (it != value_.end())
	{
		throw simplexml::XMLException("value is not integer");
	}

	INT_64 res = strtoll(value_.c_str(), NULL, 10);
	if (errno == EINVAL || errno == ERANGE)
	{
		throw simplexml::XMLException(strerror(errno));
	}
	return res;
}

bool XMLValue::AsBool() const
{
	const INT_16 bitsInbyte = 8;
	CCHAR_P trueSynonims[] = { "true", "on", "yes" };
	for (UINT_16 i = 0; i < sizeof(trueSynonims)/bitsInbyte; ++i)
	{
		if (value_ == trueSynonims[i]) { return true; }
	}
	//if (AsInteger() != 0) { return true; }
	return false;
}

W_FLOAT XMLValue::AsFloat() const
{
	std::string::const_iterator it = value_.begin();

	if (*it == '-') { ++it; }

	it = std::find_if(it, value_.end(), IsNotDigitOrPoint());

	if (it != value_.end())
	{
		throw simplexml::XMLException("value is not float");
	}

	W_FLOAT res = strtod(value_.c_str(), NULL);
	if (errno == ERANGE)
	{
		throw simplexml::XMLException(strerror(errno));
	}
	return res;
}

} // namespace simplexml

